IMPLEMENTATION MODULE ChipIO;

  IMPORT IO, Window, FIO;
  FROM Storage IMPORT ALLOCATE, DEALLOCATE;

  PROCEDURE WriteLn;
    BEGIN
      IO.WrLn;
      Done:=IO.OK
    END WriteLn;

  PROCEDURE WriteString(s : ARRAY OF CHAR);
    BEGIN
      IO.WrStr(s);
      Done:=IO.OK
    END WriteString;

  PROCEDURE WriteCard(x, n : CARDINAL);
    BEGIN
      IO.WrCard(x, n);
      Done:=IO.OK
    END WriteCard;

  PROCEDURE WriteInt(x : INTEGER; n : CARDINAL);
    BEGIN
      IO.WrInt(x, n);
      Done:=IO.OK
    END WriteInt;

  PROCEDURE WriteReal(x : REAL; m, n : CARDINAL);
    BEGIN
      IO.WrReal(x, m, n);
      Done:=IO.OK
    END WriteReal;

  PROCEDURE Write(c : CHAR);
    BEGIN
      IO.WrChar(c);
      Done:=IO.OK
    END Write;

  PROCEDURE ReadLn;
    BEGIN
      IO.RdLn;
      Done:=IO.OK
    END ReadLn;

  PROCEDURE ReadString(VAR s : ARRAY OF CHAR);
    BEGIN
      IO.RdStr(s);
      Done:=IO.OK
    END ReadString;

  PROCEDURE ReadCard(VAR x : CARDINAL);
    BEGIN
      x:=IO.RdCard();
      Done:=IO.OK
    END ReadCard;

  PROCEDURE ReadInt(VAR x : INTEGER);
    BEGIN
      x:=IO.RdInt();
      Done:=IO.OK
    END ReadInt;

  PROCEDURE ReadReal(VAR x : REAL);
    BEGIN
      x:=IO.RdReal();
      Done:=IO.OK
    END ReadReal;

  PROCEDURE Read(VAR c : CHAR);
    BEGIN
      c:=IO.RdChar();
      Done:=IO.OK
    END Read;

  PROCEDURE KeyPressed() : BOOLEAN;
    BEGIN
      Done:=TRUE;
      RETURN IO.KeyPressed()
    END KeyPressed;

  PROCEDURE ReadKey() : CHAR;
    BEGIN
      Done:=TRUE;
      RETURN IO.RdKey()
    END ReadKey;

  PROCEDURE GotoXY(x, y : CARDINAL);
    BEGIN
      Window.GotoXY(x, y)
    END GotoXY;

  PROCEDURE ClrScr;
    BEGIN
      Window.Clear
    END ClrScr;

  PROCEDURE ClrEol;
    BEGIN
      Window.ClrEol
    END ClrEol;

  PROCEDURE InsLine;
    BEGIN
      Window.InsLine
    END InsLine;

  PROCEDURE DelLine;
    BEGIN
      Window.DelLine
    END DelLine;

  PROCEDURE HighVideo;
    BEGIN
      Window.TextColor(VAL(Window.Color,15));
      Window.TextBackground(VAL(Window.Color,0))
    END HighVideo;

  PROCEDURE LowVideo;
    BEGIN
      Window.TextColor(VAL(Window.Color,7));
      Window.TextBackground(VAL(Window.Color,0))
    END LowVideo;

  PROCEDURE TextColor(c : CARDINAL);
    BEGIN
      Window.TextColor(VAL(Window.Color,c))
    END TextColor;

  PROCEDURE TextBackground(c : CARDINAL);
    BEGIN
      Window.TextBackground(VAL(Window.Color,c))
    END TextBackground;

  PROCEDURE WhereX () : CARDINAL;
    BEGIN
      RETURN Window.WhereX()
    END WhereX;

  PROCEDURE WhereY () : CARDINAL;
    BEGIN
      RETURN Window.WhereY()
    END WhereY;

  VAR Windows : ARRAY [0..255] OF Window.WinType;
      Screen : Window.WinType;

  PROCEDURE Normalize();
    VAR i : CARDINAL;
        ScrDef : Window.WinDef;
    BEGIN
      IF ScreenPtr=MaxScreen THEN
        Screen:=NIL
      ELSE
        ScrDef:=Window.FullScreenDef;
        ScrDef.Hidden:=TRUE;
        ScrDef.CursorOn:=FALSE;
        Screen:=Window.Open(ScrDef);
        Window.SnapShot;
        Window.PutOnTop(Screen);
        FOR i:=MaxScreen TO 1 BY -1 DO
          Window.PutBeneath(Windows[i-1],Windows[i]);
        END;
      END;
    END Normalize;

  PROCEDURE OpenWindow(x1, y1, x2, y2 : CARDINAL);
    VAR WD : Window.WinDef;
    BEGIN
      Normalize();
      INC(MaxScreen);
      ScreenPtr:=MaxScreen;
      WD := Window.WinDef
               (0,0,0,0,                        (* Coords *)
                Window.LightGray, Window.Black, (* Foreground/Background *)
                TRUE,                           (* Cursor On *)
                TRUE,                           (* Wrap On *)
                FALSE,                          (* Hidden Off *)
                TRUE,                           (* Frame On *)
                Window.DoubleFrame,             (* Frame *)
                Window.White, Window.Black);    (* FrameColor *)
      WD.X1:=x1-1; WD.Y1:=y1-1;
      WD.X2:=x2-1; WD.Y2:=y2-1;
      Windows[ScreenPtr] := Window.Open(WD);
      IF Screen<>NIL THEN Window.Close(Screen) END;
    END OpenWindow;

  PROCEDURE OpenWindowHeader(x1, y1, x2, y2 : CARDINAL; h : ARRAY OF CHAR);
    BEGIN
      OpenWindow(x1, y1, x2, y2);
      Window.SetTitle(Windows[ScreenPtr],h,Window.CenterUpperTitle);
    END OpenWindowHeader;

  PROCEDURE CloseWindow;
    BEGIN
      IF MaxScreen<>0
      THEN
        Normalize();
        Window.Close(Windows[MaxScreen]);
        IF Screen<>NIL THEN Window.Close(Screen) END;
        DEC(MaxScreen);
        ScreenPtr:=MaxScreen;
      END;
    END CloseWindow;

  PROCEDURE SelWindow(i : CARDINAL);
    VAR n : CARDINAL;
    BEGIN
      IF (MaxScreen<>0) AND (i<=MaxScreen)
      THEN
        Normalize();
        Window.PutOnTop(Windows[i]);
        IF Screen<>NIL THEN Window.Close(Screen) END;
        ScreenPtr:=i;
      END;
    END SelWindow;

  PROCEDURE ChangeWindow(i : CARDINAL);
    BEGIN
      IF i<=MaxScreen
      THEN
        Window.Use(Windows[i]);
        ScreenPtr:=i;
      END;
    END ChangeWindow;

  TYPE LowFile = POINTER TO FIO.File;

  PROCEDURE OpenLowFile(VAR f : LowFile; fn : ARRAY OF CHAR);
  BEGIN
    NEW(f);
    f^:=FIO.Open(fn);
    Done:=FIO.IOresult()=0;
    IF NOT Done THEN DISPOSE(f); f:=NIL; END;
  END OpenLowFile;

  PROCEDURE CreateLowFile(VAR f : LowFile; fn : ARRAY OF CHAR);
  BEGIN
    NEW(f);
    f^:=FIO.Create(fn);
    Done:=FIO.IOresult()=0;
    IF NOT Done THEN DISPOSE(f); f:=NIL; END;
  END CreateLowFile;

  PROCEDURE RenameLowFile(f1, f2 : ARRAY OF CHAR);
  BEGIN
    FIO.Rename(f1,f2);
    Done:=FIO.IOresult()=0
  END RenameLowFile;

  PROCEDURE EraseLowFile(fn : ARRAY OF CHAR);
  BEGIN
    FIO.Erase(fn);
    Done:=FIO.IOresult()=0
  END EraseLowFile;

  PROCEDURE CloseLowFile(VAR f : LowFile);
  BEGIN
    IF f=NIL THEN Done:=FALSE; RETURN; END;
    FIO.Close(f^);
    DISPOSE(f);
    Done:=FIO.IOresult()=0
  END CloseLowFile;

  PROCEDURE LongFileSize(VAR f : LowFile) : FileSizeTyp;
  VAR s:FileSizeTyp;
  BEGIN
    IF f=NIL
    THEN Done:=FALSE; RETURN 0;
    ELSE s:=FIO.Size(f^);
         Done:=FIO.IOresult()=0;
         RETURN s;
    END
  END LongFileSize;

  PROCEDURE WriteNBytes(VAR f : LowFile;
                        pos : FileSizeTyp; anz : CARDINAL;
                        VAR buffer : ARRAY OF BYTE);
    BEGIN
      IF f=NIL THEN Done:=FALSE; RETURN; END;
      FIO.Seek(f^,pos); Done:=FIO.IOresult()=0;
      IF Done THEN
        FIO.WrBin(f^,buffer,anz);
        Done:=FIO.IOresult()=0
      END (* IF *)
    END WriteNBytes;

  PROCEDURE ReadNBytes(VAR f : LowFile;
                       pos : FileSizeTyp; anz : CARDINAL;
                       VAR buffer : ARRAY OF BYTE);
  VAR n : CARDINAL;
  BEGIN
    IF f=NIL THEN Done:=FALSE; RETURN; END;
    FIO.Seek(f^,pos); Done:=FIO.IOresult()=0;
    IF Done THEN
      n:=FIO.RdBin(f^,buffer,anz);
      Done:=(FIO.IOresult()=0) AND (n=anz)
    END (* IF *)
  END ReadNBytes;

BEGIN
  FIO.IOcheck:=FALSE;
  MaxScreen:=0;
  ScreenPtr:=0;
  Windows[0]:=Window.FullScreen;
  Done:=TRUE;
END ChipIO.
