MODULE EditFile;

FROM LOWFILES IMPORT LowFile, ResetLowFile, CloseLowFile,
                     ReadNBytes, WriteNBytes, LowFileSize;

FROM ChipIO IMPORT Write, WriteString, WriteLn, WriteCard, ReadString,
                   ClrScr, GotoXY, ReadKey, ClrEol, KeyPressed,
                   ReadCard;

FROM ReadCH IMPORT ReadChar, Up, Dn, Le, Ri, Pd, Pu, It, Dl;

TYPE SHORTCARD = [0..255];

VAR FileName : ARRAY[0..64] OF CHAR;
    Result : INTEGER;
    File : LowFile;
    Size : LONGCARD;
    LastPage : CARDINAL;
    FinalByte : CARDINAL;

    Puffer : RECORD
              LastNr,
              LastByte: CARDINAL;
              Changed : BOOLEAN;
              Mem     : ARRAY[0..255] OF BYTE
             END;

PROCEDURE PutPage;
  BEGIN
    WITH Puffer DO
      IF Changed
      THEN WriteNBytes(File,LONGCARD(LastNr)*256,1+LastByte,Mem,Result)
      END;
      Changed:=FALSE
    END
  END PutPage;

PROCEDURE GetPage(Nr : CARDINAL);
  BEGIN
    WITH Puffer DO
      PutPage;
      LastNr:=Nr;
      ReadNBytes(File,LONGCARD(Nr)*256,256,Mem,Result);
    END
  END GetPage;

PROCEDURE SetByte(b : CARDINAL; h : BYTE);
  BEGIN
    WITH Puffer DO
      IF Mem[b]<>h
      THEN Mem[b]:=h;
        IF Changed
        THEN IF b>LastByte THEN LastByte:=b END
        ELSE LastByte:=b; Changed:=TRUE
        END;
      END
    END
  END SetByte;

PROCEDURE Find(p : LONGCARD; A : ARRAY OF BYTE; l : CARDINAL) : LONGCARD;
  VAR dBuffer : RECORD
                  CASE b : BOOLEAN OF
                  TRUE : b0 : ARRAY[1..2],[0..255] OF BYTE |
                  FALSE: b1 : ARRAY[0..511] OF BYTE
                  END
                END;
      i, j, r : CARDINAL;
  BEGIN
    PutPage;
    IF Size-p>512 THEN r:=512 ELSE r:=CARDINAL(Size-p) END;
    ReadNBytes(File,p,r,dBuffer.b1,Result);
    IF r>256 THEN r:=256 END;
    WITH dBuffer DO
      WHILE (r>0) AND (Result=0) DO
        FOR i:=0 TO 255 DO
          j:=0;
          WHILE (j<l) AND (A[j]=b1[i+j]) DO
            IF p+LONGCARD(i+j)>Size THEN RETURN 0 END;
            INC(j)
          END;
          IF j=l THEN RETURN p+LONGCARD(i+1) END
        END;
        b0[1]:=b0[2]; INC(p,LONGCARD(r));
        IF Size-p>256 THEN r:=256 ELSE r:=CARDINAL(Size-p) END;
        ReadNBytes(File,p+256,r,b0[2],Result)
      END;
    END;
    RETURN 0
  END Find;

PROCEDURE WriteHex(h : BYTE);
  VAR n1, n2 : SHORTCARD;

  PROCEDURE WriteNibble(n : SHORTCARD);
    BEGIN
      CASE n OF
        0..9   : Write(CHR(48+n)) |
        10..15 : Write(CHR(55+n))
       END
    END WriteNibble;

  BEGIN
    WriteNibble(SHORTCARD(h) DIV 16);
    WriteNibble(SHORTCARD(h) MOD 16);
  END WriteHex;

PROCEDURE ReadHex(VAR C : CHAR) : BOOLEAN;
  VAR n : CARDINAL;
  BEGIN
    C:=CAP(C);
    CASE C OF
    '0'..'9' : n:=ORD(C)-48 |
    'A'..'F' : n:=ORD(C)-55
    ELSE RETURN FALSE
    END;
    n:=n*16; Write(C); ReadChar(C); C:=CAP(C);
    CASE C OF
    '0'..'9' : n:=n+ORD(C)-48 |
    'A'..'F' : n:=n+ORD(C)-55
    ELSE RETURN FALSE
    END;
    Write(C); C:=CHR(n);
    RETURN TRUE
  END ReadHex;

PROCEDURE WriteASCII(h : BYTE);
  BEGIN
    IF h>=BYTE(32) THEN Write(CHR(h)) ELSE Write(".") END
  END WriteASCII;

PROCEDURE WritePage;
  VAR i,j : CARDINAL;
  BEGIN
    GotoXY(1,1); WriteString(FileName);
    WriteLn; WriteLn;
    FOR i:=0 TO 15 DO
      FOR j:=0 TO 15 DO
        IF i*16+j<FinalByte
        THEN WriteHex(Puffer.Mem[i*16+j]); Write(" ")
        ELSE WriteString("   ")
        END
      END;
      WriteString("   ");
      FOR j:=0 TO 15 DO
        IF i*16+j<FinalByte
        THEN WriteASCII(Puffer.Mem[i*16+j])
        ELSE Write(" ")
        END
      END;
      WriteLn
    END
  END WritePage;

PROCEDURE Edit;
  CONST ESC = 33C; (* ESC *)
        Fi  = 06C; (* ^F *)
        Co  = 14C; (* ^L *)
        CM  = 25C; (* ^U *)
        Sp  = 20C; (* ^P *)

  TYPE String = ARRAY[0..80] OF CHAR;

  VAR C : CHAR;
      Mode : (Hex,ASCII);
      i,j,l : CARDINAL;
      Page : CARDINAL;
      FP : LONGCARD;
      SearchStr : String;
      OverWriteMode : BOOLEAN;

  PROCEDURE ReadSearchStr;
    VAR i,b : CARDINAL;
        temp : String;

    PROCEDURE HexZiffer(C : CHAR) : BOOLEAN;
      BEGIN
        CASE CAP(C) OF
        "0".."9",
        "A".."F" : RETURN TRUE
        ELSE RETURN FALSE
        END
      END HexZiffer;

    PROCEDURE HexWert(C : CHAR) : CARDINAL;
      BEGIN
        CASE CAP(C) OF
        "0".."9" : RETURN ORD(C)-ORD("0") |
        "A".."F" : RETURN ORD(CAP(C))+10-ORD("A")
        ELSE (* Fehler *)
        END
      END HexWert;

    BEGIN
      GotoXY(1,2); WriteString("Suchen: ");
      ReadString(SearchStr);
      IF Mode=ASCII
      THEN l:=0; WHILE SearchStr[l]<>0C DO INC(l) END
      ELSE
        l:=0; i:=0; temp:=SearchStr; b:=0;
        REPEAT
          IF HexZiffer(temp[i])
          THEN
             b:=HexWert(temp[i]); INC(i);
             IF HexZiffer(temp[i])
             THEN b:=b*16+HexWert(temp[i]); SearchStr[l]:=CHR(b); INC(l)
             END
          END;
          IF temp[i]<>0C THEN INC(i) END
        UNTIL temp[i]=0C
      END
    END ReadSearchStr;

  PROCEDURE Show;
    BEGIN
      GetPage(Page);
      IF Page=LastPage
      THEN FinalByte:=CARDINAL(Size MOD 256)
      ELSE FinalByte:=256
      END;
      GotoXY(1,1); WritePage
    END Show;

  PROCEDURE SetCursor;
    BEGIN
      GotoXY(1,2); ClrEol;
      IF j*16+i>=FinalByte-1
      THEN j:=(FinalByte-1) DIV 16; i:=(FinalByte-1) MOD 16
      END;
      GotoXY(40,1); WriteString('Seite: '); WriteCard(Page,4);
      GotoXY(52,1); WriteString('Byte: '); WriteCard(j*16+i,3);
      IF Mode=Hex THEN GotoXY(1+i*3,j+3) ELSE GotoXY(52+i,j+3) END
    END SetCursor;

  PROCEDURE ShowByte;
    BEGIN
      GotoXY(1+3*i,j+3); WriteHex(Puffer.Mem[j*16+i]);
      GotoXY(52+i, j+3); WriteASCII(Puffer.Mem[j*16+i]);
    END ShowByte;

  PROCEDURE Rechts;
    BEGIN
      IF i<15 THEN INC(i)
      ELSIF j<15 THEN INC(j); i:=0
      ELSIF Page<LastPage THEN INC(Page); i:=0; j:=0; Show
      END
    END Rechts;

  PROCEDURE Links;
    BEGIN
      IF i>0 THEN DEC(i)
      ELSIF j>0 THEN DEC(j); i:=15
      ELSIF Page>0 THEN DEC(Page); i:=15; j:=15; Show
      END
    END Links;

  BEGIN
    Size:=LowFileSize(File,Result); SearchStr:=""; l:=0;
    LastPage:=CARDINAL((Size+255) DIV 256)-1;
    OverWriteMode:=FALSE;
    ClrScr; Page:=0; Show; i:=0; j:=0; Mode:=Hex;
    GotoXY(1,20);
    WriteString("^F : Suchen            ^L : Suche fortsetzen"); WriteLn;
    WriteString("^P : Seite setzen      ^U : Schreibschutz umschalten");
    WriteLn;
    WriteString("^V : Modus umschalten"); WriteLn; WriteLn;
    WriteString("ESC: Programm beenden");
    GotoXY(65,1); WriteString("Schreibschutz");
    REPEAT
      SetCursor;
      ReadChar(C);
      CASE C OF
      Pu : IF Page>0 THEN DEC(Page); Show END |
      Pd : IF Page<LastPage THEN INC(Page); Show END |
      It : IF Mode=Hex THEN Mode:=ASCII ELSE Mode:=Hex END |
      Le : Links |
      Ri : Rechts |
      Up : IF j>0 THEN DEC(j)
           ELSIF Page>0 THEN DEC(Page); j:=15; Show
           END |
      Dn : IF j<15 THEN INC(j)
           ELSIF Page<LastPage THEN INC(Page); j:=0; Show
           END |
      Fi : ReadSearchStr;
           FP:=Find(0,SearchStr,l);
           IF FP>0
           THEN
             DEC(FP); Page:=CARDINAL(FP DIV 256); FP:=FP MOD 256;
             j:=CARDINAL(FP DIV 16); i:=CARDINAL(FP MOD 16); Show
           ELSE
             GotoXY(1,2);
             WriteString("Nicht gefunden: "); WriteString(SearchStr); ClrEol;
             REPEAT UNTIL KeyPressed()
           END; |
      Co : FP:=Find(1+256*LONGCARD(Page)+LONGCARD(j*16+i),SearchStr,l);
           IF FP>0
           THEN
             DEC(FP); Page:=CARDINAL(FP DIV 256); FP:=FP MOD 256;
             j:=CARDINAL(FP DIV 16); i:=CARDINAL(FP MOD 16); Show
           ELSE
             GotoXY(1,2);
             WriteString("Nicht gefunden: "); WriteString(SearchStr); ClrEol;
             REPEAT UNTIL KeyPressed()
           END; |
      Sp  : GotoXY(1,2); WriteString("Seite: ");
            ReadCard(Page);
            IF Page>LastPage THEN Page:=LastPage END;
            GotoXY(1,2); ClrEol; Show |
      ESC :     |
      CM  : OverWriteMode:=NOT OverWriteMode;
            GotoXY(65,1);
            IF OverWriteMode
            THEN WriteString("             ")
            ELSE WriteString("Schreibschutz")
            END |
      ELSE IF OverWriteMode THEN
          IF Mode=ASCII
          THEN IF C>=' '
               THEN Write(C); SetByte(j*16+i,BYTE(C)); ShowByte; Rechts
               END
          ELSE IF ReadHex(C)
               THEN SetByte(j*16+i,BYTE(C)); C:=0C; Rechts
               ELSE Show
               END
          END
        END
      END
    UNTIL C=ESC;
    ClrScr
  END Edit;

BEGIN
  Puffer.Changed:=FALSE;
  WriteString("Filename: ");
  ReadString(FileName);
  WriteLn;
  ResetLowFile(File,FileName,Result);
  IF Result=0
  THEN Edit; PutPage; CloseLowFile(File,Result)
  ELSE WriteString("Keine Datei")
  END;
  WriteLn;
END EditFile.
