IMPLEMENTATION MODULE BigArrays;

FROM SYSTEM IMPORT TSIZE, ADDRESS;
FROM Storage IMPORT Available, ALLOCATE, DEALLOCATE;

TYPE PtrArray = ARRAY[0..MAX(INTEGER) DIV TSIZE(ADDRESS)] OF ADDRESS;
     BigArrayType = RECORD
                      LastIndex : CARDINAL;
                      EntrySize : CARDINAL;
                      Array : POINTER TO PtrArray
                    END;
     BigArray = POINTER TO BigArrayType;

PROCEDURE Copy(A, B : ADDRESS; Size : CARDINAL);
  (* Damit werden Size Bytes von A^ nach B^ kopiert. In den meisten
     Modula-2-Implementierungen gibt es hierfr eine schnelle
     Bibliotheksroutine, die selbstverstndlich verwendet werden
     sollte. *)

  TYPE Memory = ARRAY[0..MAX(INTEGER)] OF BYTE;
  VAR  X,Y : POINTER TO Memory;
       i : CARDINAL;
  BEGIN
    X:=A; Y:=B;
    FOR i:=1 TO Size DO Y^[i-1]:=X^[i-1] END
  END Copy;

PROCEDURE InitArray(VAR BA : BigArray; Max, Size : CARDINAL) : BOOLEAN;
  VAR i : CARDINAL;
  BEGIN
    IF NOT Available(TSIZE(BigArrayType)+(Max+1)*TSIZE(ADDRESS))
    THEN BA:=NIL; RETURN FALSE
    ELSE
      ALLOCATE(BA,TSIZE(BigArrayType));
      WITH BA^ DO
        LastIndex:=Max;
        EntrySize:=Size;
        ALLOCATE(Array,(Max+1)*TSIZE(ADDRESS));
        FOR i:=0 TO LastIndex DO Array^[i]:=NIL END;
        RETURN TRUE
      END
    END
  END InitArray;

PROCEDURE PutItem(BA : BigArray; Index : CARDINAL; Item : ADDRESS) : BOOLEAN;
  BEGIN
    IF (BA=NIL) OR (BA^.LastIndex<Index)
    THEN RETURN FALSE
    ELSE
      WITH BA^ DO
        IF Array^[Index]=NIL
        THEN
          IF NOT Available(EntrySize)
          THEN RETURN FALSE
          ELSE ALLOCATE(Array^[Index],EntrySize)
          END
        END;
        Copy(Item,Array^[Index],EntrySize);
        RETURN TRUE
      END
    END
  END PutItem;

PROCEDURE GetItem(BA : BigArray; Index : CARDINAL; Item : ADDRESS) : BOOLEAN;
  BEGIN
    IF (BA=NIL) OR (BA^.LastIndex<Index) OR (BA^.Array^[Index]=NIL)
    THEN RETURN FALSE
    ELSE
      WITH BA^ DO
        Copy(Array^[Index],Item,EntrySize);
        RETURN TRUE
      END
    END
  END GetItem;

PROCEDURE ExitArray(VAR BA : BigArray);
  VAR i : CARDINAL;
  BEGIN
    IF BA<>NIL THEN
      WITH BA^ DO
        FOR i:=0 TO LastIndex DO
          IF Array^[i]<>NIL THEN DEALLOCATE(Array^[i],EntrySize) END;
        END;
        DEALLOCATE(Array,(LastIndex+1)*TSIZE(ADDRESS))
      END;
      DEALLOCATE(BA,TSIZE(BigArrayType))
    END
  END ExitArray;

PROCEDURE SortArray(BA : BigArray; less : Relation; Last : CARDINAL);

  PROCEDURE QSort(f,l : INTEGER);
    VAR i,j : INTEGER;
        m, t : ADDRESS;
    BEGIN
      WITH BA^ DO
        i:=f; j:=l; m:=Array^[(i+j) DIV 2];
        WHILE i<j DO
          WHILE less(Array^[i],m) DO INC(i) END;
          WHILE less(m,Array^[j]) DO DEC(j) END;
          IF i<=j THEN
            t:=Array^[i]; Array^[i]:=Array^[j]; Array^[j]:=t;
            IF m=Array^[i] THEN m:=Array^[j]
            ELSIF m=Array^[j] THEN m:=Array^[i]
            END;
            INC(i); DEC(j)
          END
        END;
        IF i<l THEN QSort(i,l) END;
        IF f<j THEN QSort(f,j) END
      END
    END QSort;

  BEGIN
    QSort(0,Last)
  END SortArray;

END BigArrays.