IMPLEMENTATION MODULE BinTrees;

FROM SYSTEM IMPORT TSIZE;
FROM Storage IMPORT Available, ALLOCATE, DEALLOCATE;

TYPE Tree = POINTER TO Node;
     Node = RECORD
              MPtr : ADDRESS;
              MSize : CARDINAL;
              left, right : Tree
            END;

     BinTreeType = RECORD
                     T : Tree;
                     less : LessFunction
                   END;
     BinTree = POINTER TO BinTreeType;

PROCEDURE Init(VAR B : BinTree; LF : LessFunction) : BOOLEAN;
  BEGIN
    IF NOT Available(TSIZE(BinTreeType))
    THEN RETURN FALSE
    ELSE
      ALLOCATE(B,TSIZE(BinTreeType));
      WITH B^ DO T:=NIL; less:=LF END;
      RETURN TRUE
    END
  END Init;

PROCEDURE Exit(VAR B : BinTree);

  PROCEDURE DelTree(VAR T : Tree);
    BEGIN
      IF T<>NIL THEN
        WITH T^ DO
          DelTree(left); DelTree(right);
          DEALLOCATE(MPtr,MSize)
        END;
        DEALLOCATE(T,TSIZE(Node))
      END
    END DelTree;

  BEGIN
    DelTree(B^.T);
    DEALLOCATE(B,TSIZE(BinTreeType))
  END Exit;

PROCEDURE Copy(A, B : ADDRESS; Size : CARDINAL);
  TYPE Memory = ARRAY[0..MAX(INTEGER)] OF BYTE;

  PROCEDURE C(VAR X,Y : Memory);
    VAR i : CARDINAL;
    BEGIN
      FOR i:=0 TO Size-1 DO
        Y[i]:=X[i]
      END
    END C;

  BEGIN
    C(Memory(A^),Memory(B^))
  END Copy;

PROCEDURE Insert(B : BinTree; Item : ADDRESS; Size : CARDINAL): BOOLEAN;

  PROCEDURE InsNode(VAR T : Tree);
    BEGIN
      IF T=NIL
      THEN
        ALLOCATE(T,TSIZE(Node));
        WITH T^ DO
          left:=NIL; right:=NIL;
          MSize:=Size;
          ALLOCATE(MPtr,Size);
          Copy(Item,MPtr,Size)
        END
      ELSIF B^.less(Item,T^.MPtr) THEN InsNode(T^.left)
      ELSE InsNode(T^.right)
      END
    END InsNode;

  BEGIN
    IF NOT Available(TSIZE(Node)+Size)
    THEN RETURN FALSE
    ELSE InsNode(B^.T); RETURN TRUE
    END
  END Insert;

PROCEDURE Delete(B : BinTree; Item : ADDRESS) : BOOLEAN;

  PROCEDURE DelNode(VAR T : Tree) : BOOLEAN;
    VAR p, q, swap : Tree;

    PROCEDURE free;
      BEGIN
        WITH p^ DO DEALLOCATE(MPtr,MSize) END;
        DEALLOCATE(p,TSIZE(Node))
      END free;

    BEGIN
      IF T<>NIL THEN
        IF B^.less(Item,T^.MPtr) THEN RETURN DelNode(T^.left)
        ELSIF B^.less(T^.MPtr,Item) THEN RETURN DelNode(T^.right)
        ELSE
          p:=T;
          IF T^.left=NIL THEN T:=T^.right; free; RETURN TRUE
          ELSIF T^.right=NIL THEN T:=T^.left; free; RETURN TRUE
          ELSE
            p:=T^.left;
            WHILE p^.right<>NIL DO p:=p^.right END;
            swap:=T^.MPtr; T^.MPtr:=p^.MPtr; p^.MPtr:=swap;
            RETURN DelNode(T^.left)
          END
        END
      ELSE RETURN FALSE
      END
    END DelNode;

  BEGIN
    RETURN DelNode(B^.T);
  END Delete;

PROCEDURE Member(B : BinTree; Item : ADDRESS) : BOOLEAN;
  VAR p : Tree;
  BEGIN
    WITH B^ DO
      p:=T;
      WHILE p<>NIL DO
        IF less(Item,p^.MPtr) THEN p:=p^.left
        ELSIF less(p^.MPtr,Item) THEN p:=p^.right
        ELSE Copy(p^.MPtr,Item,p^.MSize); RETURN TRUE
        END
      END;
      RETURN FALSE
    END
  END Member;


END BinTrees.