IMPLEMENTATION MODULE HASH;

FROM SYSTEM IMPORT TSIZE, BYTE;
FROM Storage IMPORT Available, ALLOCATE, DEALLOCATE;
  (* Falls Available nicht vorhanden ist, mu eine andere
     Funktion zur Heapberwachung importiert werden. *)
FROM InOut IMPORT WriteString, WriteLn;

TYPE List = POINTER TO Node;

     Node = RECORD
              MItem : ADDRESS;
              MSize : CARDINAL;
              next : List
            END;

     TableType = ARRAY[0..MAX(INTEGER) DIV TSIZE(ADDRESS)] OF List;

     TableRec =  RECORD
                   Equal : EqualFunc;
                   Hash  : HashFunc;
                   MaxTable : CARDINAL;
                   Tab : POINTER TO TableType
                 END;


     Table = POINTER TO TableRec;

PROCEDURE TestInit(T : Table) : BOOLEAN;
  BEGIN
    IF T=NIL
    THEN
      WriteString("Hashmodul nicht initialisiert!");
      WriteLn;
      RETURN FALSE
    ELSE RETURN TRUE
    END
  END TestInit;

PROCEDURE Init(VAR T : Table; Max : CARDINAL; HF : HashFunc; Eq : EqualFunc)
           : BOOLEAN;
  VAR i : CARDINAL;
  BEGIN
    IF NOT Available(TSIZE(TableRec)+(Max-1)*TSIZE(ADDRESS))
    THEN RETURN FALSE
    ELSE
      ALLOCATE(T,TSIZE(TableRec));
      WITH T^ DO
        ALLOCATE(Tab,(Max-1)*TSIZE(ADDRESS));
        FOR i:=0 TO Max-1 DO Tab^[i]:=NIL END;
        Hash:=HF;
        Equal:=Eq;
        MaxTable:=Max
      END;
      RETURN TRUE
    END
  END Init;

PROCEDURE Find(T : Table; Item : ADDRESS) : ADDRESS;
  (* Diese Funktion gibt einen Zeiger auf den Listeneintrag zurck,
     der den gesuchten Eintrag enthlt. NIL bedeutet, da der
     Eintrag nicht in der Tabelle ist. *)
  VAR Index : CARDINAL;
      p : List;
  BEGIN
    IF TestInit(T) THEN
      WITH T^ DO
        Index:=Hash(Item) MOD MaxTable;
        p:=Tab^[Index];
        WHILE (p<>NIL) AND NOT Equal(p^.MItem,Item) DO p:=p^.next END;
        RETURN p
      END
    END
  END Find;

PROCEDURE Copy(A, B : ADDRESS; Size : CARDINAL);
  (* Damit werden Size Bytes von A^ nach B^ kopiert. In den meisten
     Modula-2-Implementierungen gibt es hierfr eine schnelle
     Bibliotheksroutine, die selbstverstndlich verwendet werden
     sollte. *)

  TYPE Memory = ARRAY[0..MAX(INTEGER)] OF BYTE;
  VAR  X,Y : POINTER TO Memory;
       i : CARDINAL;
  BEGIN
    X:=A; Y:=B;
    FOR i:=1 TO Size DO Y^[i-1]:=X^[i-1] END
  END Copy;

PROCEDURE Member(T : Table; Item : ADDRESS) : BOOLEAN;
  VAR p : List;
  BEGIN
    IF TestInit(T) THEN
      p:=Find(T,Item);
      IF p=NIL
      THEN RETURN FALSE
      ELSE
        Copy(p^.MItem,Item,p^.MSize);
        RETURN TRUE
      END
    END
  END Member;

PROCEDURE Insert(T : Table; Item : ADDRESS; Size : CARDINAL) : BOOLEAN;

  PROCEDURE InsList(VAR L : List);
    VAR p : List;
    BEGIN
      ALLOCATE(p,TSIZE(Node));
      WITH p^ DO
        next:=L;
        MSize:=Size;
        ALLOCATE(MItem,MSize);
        Copy(Item,MItem,MSize)
      END;
      L:=p;
    END InsList;

  BEGIN (* Insert *)
    IF TestInit(T) THEN
      IF Available(Size+TSIZE(Node))
      THEN
        WITH T^ DO
          InsList(Tab^[Hash(Item) MOD MaxTable]);
          RETURN TRUE
        END
      ELSE RETURN FALSE
      END
    END
  END Insert;

PROCEDURE Delete(T : Table; Item : ADDRESS) : BOOLEAN;
  (* Delete arbeitet mit einem rekursiven Algorithmus. *)
  VAR p : List;

  PROCEDURE Free; (* Gibt den Speicherplatz eines Listeneintrags frei *)
    BEGIN
      WITH p^ DO DEALLOCATE(MItem,MSize) END;
      DEALLOCATE(p,TSIZE(Node))
    END Free;

  PROCEDURE DelList(VAR L : List);
    BEGIN
      IF L=p
      THEN L:=p^.next; Free
      ELSIF L^.next=p
      THEN L^.next:=p^.next; Free
      ELSIF L<>NIL THEN DelList(L^.next)
      END
    END DelList;

  BEGIN
    IF TestInit(T) THEN
      p:=Find(T,Item);
      IF p=NIL
      THEN RETURN FALSE
      ELSE
        WITH T^ DO
          DelList(Tab^[Hash(Item) MOD MaxTable]);
          RETURN TRUE
        END
      END
    END
  END Delete;

PROCEDURE Manipu(T : Table; Item : ADDRESS; Op : ItemOp);
  VAR p : List;
  BEGIN
    IF TestInit(T) THEN
      p:=Find(T,Item);
      IF p<>NIL THEN
        Op(p^.MItem);
      END
    END
  END Manipu;

PROCEDURE WriteAll(T : Table; Op : ItemOp);
  VAR i : CARDINAL;
      p : List;
  BEGIN
    IF TestInit(T) THEN
      WITH T^ DO
        FOR i:=0 TO MaxTable-1 DO
          p:=Tab^[i];
          WHILE p<>NIL DO
            Op(p^.MItem);
            p:=p^.next
          END
        END
      END
    END
  END WriteAll;

PROCEDURE Exit(VAR T : Table);
  (* Gibt den belegten Speicherplatz frei und deinitialisiert das Modul *)
  VAR i : CARDINAL;
  BEGIN
    IF TestInit(T) THEN
      WITH T^ DO
        FOR i:=0 TO MaxTable-1 DO
          WHILE (Tab^[i]<>NIL) AND Delete(T,Tab^[i]^.MItem)
          DO (* Nothing *)
          END
        END;
        DEALLOCATE(Tab,(MaxTable-1)*TSIZE(ADDRESS))
      END;
      DEALLOCATE(T,TSIZE(TableRec));
      T:=NIL
    END;
  END Exit;

PROCEDURE TableFree(T : Table):CARDINAL;
  VAR i, s : CARDINAL;
  BEGIN
    IF TestInit(T) THEN
      WITH T^ DO
        s:=0;
        FOR i:=0 TO MaxTable-1 DO
          IF Tab^[i]=NIL THEN INC(s) END
        END
      END;
      RETURN s
    END
  END TableFree;

PROCEDURE Count(T : Table):CARDINAL;
  VAR i, s : CARDINAL;
      p : List;
  BEGIN
    IF TestInit(T) THEN
      WITH T^ DO
        s:=0;
        FOR i:=0 TO MaxTable-1 DO
          p:=Tab^[i];
          WHILE p<>NIL DO INC(s); p:=p^.next END
        END
      END;
      RETURN s
    END
  END Count;

END HASH.