MODULE Lexikon;  (* Demonstration fr das Hash-Modul *)

FROM HASH IMPORT Table, Init, Exit, Member, Insert, Count, WriteAll;

FROM InOut IMPORT WriteString, WriteLn, Write, Read,
                  OpenInput, CloseInput, OpenOutput,
                  CloseOutput, Done, EOL;

FROM Terminal IMPORT BusyRead;

FROM SYSTEM IMPORT ADDRESS, ADR, TSIZE;

TYPE Eintrag = RECORD
                  Suchbegriff : ARRAY[0..20] OF CHAR;
                  Erklaerung : ARRAY[0..60] OF CHAR;
                END;

VAR  LexGroesse : CARDINAL;
     Eintraege : Table;

PROCEDURE LiesZeile(VAR S : ARRAY OF CHAR);
  VAR C : CHAR;
      l : CARDINAL;
  BEGIN
    l:=0;
    REPEAT
      BusyRead(C);
      IF C = 10C
      THEN
        IF l>0 THEN DEC(l); Write(C); Write(" "); Write(C) END
      ELSIF C>=" " THEN S[l]:=C; Write(C); INC(l)
      END
    UNTIL (l=HIGH(S)) OR (C=15C);
    S[l]:=0C
  END LiesZeile;

PROCEDURE StringEqual(VAR X, Y : ARRAY OF CHAR) : BOOLEAN;
  VAR i : CARDINAL;
  BEGIN
    i:=0;
    LOOP
      IF (i>HIGH(X)) OR (i>HIGH(Y)) THEN RETURN HIGH(X)=HIGH(Y)
      ELSIF CAP(X[i])<>CAP(Y[i])
      THEN RETURN FALSE
      ELSIF X[i]=0C THEN RETURN TRUE
      ELSE INC(i)
      END
    END
  END StringEqual;

PROCEDURE EintragGleich(t1, t2 : ADDRESS) : BOOLEAN;
  VAR E1, E2 : POINTER TO Eintrag;
  BEGIN
    E1:=t1; E2:=t2;
    RETURN StringEqual(E1^.Suchbegriff,E2^.Suchbegriff)
  END EintragGleich;

PROCEDURE HashFunktion(t : ADDRESS) : CARDINAL;
  VAR S : CARDINAL;
      i : CARDINAL;
      E : POINTER TO Eintrag;
  BEGIN
    E:=t; i:=0; S:=0;
    WHILE E^.Suchbegriff[i]<>0C DO
      INC(S,ORD(CAP(E^.Suchbegriff[i])));
      INC(i);
    END;
    RETURN S
  END HashFunktion;

PROCEDURE LiesLexikon;
  VAR E : Eintrag;

  PROCEDURE LiesZeile(VAR S : ARRAY OF CHAR);
    VAR l : CARDINAL;
        C : CHAR;
    BEGIN
      l:=0;
      WHILE Done AND (l<HIGH(S)) DO
        Read(C);
        IF C=EOL THEN S[l]:=0C; RETURN END;
        IF C>=" " THEN S[l]:=C; INC(l) END
      END;
      S[l]:=0C
    END LiesZeile;

  BEGIN
    WriteString("Lesen eines Lexikons"); WriteLn;
    OpenInput("");
    IF Init(Eintraege,1000,HashFunktion,EintragGleich)
    THEN
      WHILE Done DO
        LiesZeile(E.Suchbegriff);
        IF E.Suchbegriff[0]<>0C THEN
          LiesZeile(E.Erklaerung);
          IF NOT Insert(Eintraege,ADR(E),TSIZE(Eintrag))
          THEN WriteString("Kein Speicher!"); WriteLn;
          END
        END
      END
    ELSE WriteString("Kein Speicher!"); WriteLn; HALT
    END;
    CloseInput
  END LiesLexikon;

PROCEDURE SchreibEintrag(t : ADDRESS);
  VAR E : POINTER TO Eintrag;
  BEGIN
    E:=t;
    WITH E^ DO
      WriteString(Suchbegriff); WriteLn;
      WriteString(Erklaerung); WriteLn
    END
  END SchreibEintrag;

PROCEDURE SchreibLexikon;
  BEGIN
    WriteString("Das Lexikon wird abgelegt"); WriteLn;
    OpenOutput("");
    IF Done THEN WriteAll(Eintraege,SchreibEintrag) END;
    CloseOutput
  END SchreibLexikon;

PROCEDURE BearbeiteLexikon;
  VAR E : Eintrag;
      Merker : Table;

  PROCEDURE Buchstabe(C : CHAR) : BOOLEAN;
    BEGIN
      CASE CAP(C) OF
      "A".."Z","-",
      "0".."9",
      "","","",
      "","","","" : RETURN TRUE
      ELSE RETURN FALSE
      END
    END Buchstabe;

  PROCEDURE BearbeiteEintrag(E : Eintrag);
    VAR l, k : CARDINAL; E0 : Eintrag;
    BEGIN
      WITH E DO
        IF NOT Member(Merker,ADR(Suchbegriff)) THEN
          WriteString(Suchbegriff); WriteString(": ");
          WriteString(Erklaerung); WriteLn;
          IF Insert(Merker,ADR(Suchbegriff),SIZE(Suchbegriff)) THEN
            l:=0;
            WHILE Erklaerung[l]<>0C DO
              IF Erklaerung[l]=">"
              THEN
                k:=0; INC(l);
                WHILE (Erklaerung[l]<>0C) AND Buchstabe(Erklaerung[l]) DO
                  E0.Suchbegriff[k]:=Erklaerung[l];
                  INC(l); INC(k)
                END;
                E0.Suchbegriff[k]:=0C;
                IF Member(Eintraege,ADR(E0)) THEN BearbeiteEintrag(E0) END
              ELSE INC(l)
              END
            END
          END
        END
      END
    END BearbeiteEintrag;


  BEGIN
    LOOP
      WriteLn; WriteLn;
      WriteString("Suchbegriff: ");
      LiesZeile(E.Suchbegriff);
      IF E.Suchbegriff[0]=0C THEN EXIT END;
      WriteLn;
      IF Member(Eintraege,ADR(E)) AND
         Init(Merker,20,HashFunktion,EintragGleich)
      THEN
        BearbeiteEintrag(E);
        Exit(Merker)
      ELSE
        WriteString("Erklrung:   ");
        LiesZeile(E.Erklaerung);
        IF E.Erklaerung[0]<>0C THEN
          IF NOT Insert(Eintraege,ADR(E),TSIZE(Eintrag))
          THEN WriteString("-- kein Speicher mehr!");
          END
        END
      END
    END;
    WriteLn; WriteLn
  END BearbeiteLexikon;

BEGIN
  LiesLexikon;
  LexGroesse:=Count(Eintraege);
  BearbeiteLexikon;
  IF LexGroesse<>Count(Eintraege) THEN SchreibLexikon END
END Lexikon.

